/*
   NAME:	Frame.h
   AUTHOR:	Ole Gunnar Westgaard TFU <olew@dina>
   DATE:	Jul 12 1996
   COMMENT:	FRAME figlet for drawing frames
*/
#ifndef FRAMEH
#define FRAMEH

#include "otools.h"

#define bIN  0x80 /* otherwise bOUT (bNIL not defined) */
#define fIN  0x08 /* otherwise fOUT (fNIL not defined) */

#define bSNG 0x10 /* single 1 pixel border (bd, bIN gives sd) */
#define bDBL 0x20 /* double 1 pixel border (bd, bIN gives sd) 1pix between */
#define bSHD 0x30 /* shadow = single + 3 pixel shadow (bd + bd / sd) */
#define bTGL 0x40 /* tangled, as double, but boxes 'collide' (bd, bIN=other) */
#define bTHK 0x50 /* thick border 2 pixels (bd) (bIN gives sd color) */
#define bTWO 0x60 /* as tangled, but one box (hl) other (sd) - see bIN */
#define bTSD 0x70 /* Topshadow, as bSHD, but on top */

#define fSNG 0x01 /* Single 3d */
#define fDBL 0x02 /* Double 3d */
#define fTRP 0x03 /* Triple 3d  */
#define fWID 0x04 /* Wide 3d (down 2 pixel - up width - down 2 pixel) */
#define fPLT 0x05 /* Plateau (up 2 pixel - width level) */
#define fSNK 0x06 /* Sink (up 2pix - width level - down 2 pix) */
#define fFLR 0x07 /* Floor (up wide - 3pix level - down wide) */

// method:
#define frVERFIRST 0x20 /* Go vertical first, not horizontal */
#define frLEFT     0x40 /* Go left, not right */
#define frUP       0x80 /* Go up, not down */

class FRAME:public FIGLET {
  /// Add a frame to your window, you can stack figlets within the frame.
  /// The frame will automatically resize and position the figlets if you
  /// want it to. You can also manipulate the way that the stacking is done
  /// and how the frame should look:

  //> Definitions you may combine in your frames 'look' (see creators)
  //>
  //> #define bIN  0x80 /* otherwise bOUT (bNIL not defined) */
  //> #define fIN  0x08 /* otherwise fOUT (fNIL not defined) */
  //> #define bSNG 0x10 /* single 1 pixel border (bd, bIN gives sd) */
  //> #define bDBL 0x20 /* double 1 pixel border (bd, bIN gives sd) 1pix between */
  //> #define bSHD 0x30 /* shadow = single + 3 pixel shadow (bd + bd / sd) */
  //> #define bTGL 0x40 /* tangled, as double, but boxes 'collide' (bd, bIN=other) */
  //> #define bTHK 0x50 /* thick border 2 pixels (bd) (bIN gives sd color) */
  //> #define bTWO 0x60 /* as tangled, but one box (hl) other (sd) - see bIN */
  //> #define bTSD 0x70 /* Topshadow, as bSHD, but on top */
  //> #define fSNG 0x01 /* Single 3d */
  //> #define fDBL 0x02 /* Double 3d */
  //> #define fTRP 0x03 /* Triple 3d  */
  //> #define fWID 0x04 /* Wide 3d (down 2 pixel - up width - down 2 pixel) */
  //> #define fPLT 0x05 /* Plateau (up 2 pixel - width level) */
  //> #define fSNK 0x06 /* Sink (up 2pix - width level - down 2 pix) */
  //> #define fFLR 0x07 /* Floor (up wide - 3pix level - down wide) */


protected:
  COLORSET *cset;
  PEN *pen;
  LIST figlets;
  uchar look,thick,method,space;        // Frametype,thickness, stacking method
  int bd,sd,bg,hl,dk,                                          // Color indexes
    we,he,cx,cy,                                  // Stacking size and position
    ixa,iya,ixb,iyb;                                     // Stacking frame size
  float width,height;                         // Figlet widhin - width & height
public:
  FRAME(WINDOW*,int,int,int,int,uchar,uchar,int);
  //. (window,xa,ya,xb,yb,type,width,index)
  FRAME(WINDOW*,COLORSET*,int,int,int,int,uchar,uchar,int,int,int,int,int);
  //. (window,colorset,xa,ya,xb,yb,type,width,bd,sd,bg,hl,dk)
  virtual ~FRAME(void);
  //. Removing the frame, will rerefresh linked figlets, giving them back to
  //. The window that the frame did belong to.
  virtual void Refresh(void);
  //. Refresh the frame, and figlets within
  virtual char Resize(int,int,int,int);
  //. Will actually STACK all figlets
  virtual char Add(FIGLET*);
  //. Add a figlet into the frame, the Figlet will be resized if a stacking
  //. have been specified.
  virtual char Del(FIGLET*);
  //. Delete the figlet from the frame, the figlet will be reassigned to the
  //. window that the frame belongs to.
  virtual void Stacking(int,int,int,int,uchar,uchar);
  //. Width,height in elements, the size of the figlet array in the frame.
  //. Xpos,ypos in array position, beginning at (1,1) up to (Widht,height) to
  //. start at. Method, a combination of the following:
  //> #define frVERFIRST 0x20 /* Go vertical first, not horizontal */
  //> #define frLEFT     0x40 /* Go left, not right */
  //> #define frUP       0x80 /* Go up, not down */
  //. Space, the pixelspace between each figlet.
  virtual void Stack(void);
  //. Stack all figlets now (unsupported)
  virtual void StackSingle(FIGLET*);
  //. Stack a single figlet from method and Stacking options.
  virtual void Empty(void);
  //. Empty the frame
};

#endif /* FRAMEH */
