/*
   NAME:	input.cc
   AUTHOR:	Ole Gunnar Westgaard TFU <olew@dina>
   DATE:	Jul 16 1996
   COMMENT:	INPUT figlet, keyboard input
*/

#include "input.h"

INPUT::INPUT(WINDOW *w,COLORSET *cs,PEN *pn,int a,int b,int c,int d,
	     char *t,uchar o,ELEMENT *e,char(*cb)(INPUT*,char*))
  :FIGLET(w,a,b,c,d)
{
  set = cs ? cs : window->cset();
  pen = pn ? pn : window->dpen();
  string = NULL;
  stringsize = 0;
  options = o;
  cb_element = e;
  cb_function = cb;
  String(t);
  window->Add(this);
  if (options & inCLICKTOFOCUS)
    window->Activate(this);
  else
    window->Motion(this);
}

void INPUT::String(char *t)
{
  if (string) delete string;
  if (!t) {
    stringsize=1024;
    string = new char[stringsize];
    strcpy(string,"");
  } else {
    stringsize=1024+strlen(t);
    string = new char[stringsize];
    strcpy(string,t);
  }
  stringstart = 0;
  cursor = strlen(string);
}

INPUT::~INPUT(void)
{
  if (string) delete string;
  if (options & inFOCUS) window->DeFocus(this);
  if (options & inCLICKTOFOCUS)
    window->DeActivate(this);
  else
    window->DeMotion(this);
  window->Del(this);
}

void INPUT::DrawFocus(void)
{
  if (options & inFOCUS)
    pen->Box3d(xa,ya,xb,yb,(options & 0x0F)+1,
	       (*set)[ocHIGHLIGHT],(*set)[ocSHADOW]);
  else
    pen->Box3d(xa,ya,xb,yb,(options & 0x0F)+1,
	       (*set)[ocSHADOW],(*set)[ocHIGHLIGHT]);
}

void INPUT::Refresh(void)
{
  // Border
  DrawFocus();
  // Text
  DrawText();
}

void INPUT::DrawText(void)
{
  int w = (options & 0x0F)+1;

  // check if cursor fits within figlet
  if (stringstart > cursor)
    // No - adjust the stringstart to fit with the cursor (!)
    stringstart = cursor;
  else {
    if (string[cursor] == '\0') {
      // cursor is at end of string...
      string[cursor] = ' ';
      string[cursor+1] = '\0';
      while (pen->Width(string+stringstart) > (xb-xa-3-w*2)) stringstart++;
      string[cursor] = '\0';
    } else {
      // cursor within string...
      char c = string[cursor+1];
      string[cursor+1] = '\0';
      while (pen->Width(string+stringstart) > (xb-xa-3-w*2)) stringstart++;
      string[cursor+1] = c;
    }
  }

  // Background
  if (options & inDARKBACKGROUND)
    pen->Foreground((*set)[ocDARK]);
  else
    pen->Foreground((*set)[ocBACKGROUND]);
  pen->Fbox(xa+w,ya+w,xb-w,yb-w);

  // Draw the text and cursor
  char m[2] = { string[cursor],'\0' };
  string[cursor] = '\0';
  int x = xa + w + 2,
      y = ya+(yb-ya+1-pen->FontHeight())/2 + pen->FontHeight();

  pen->Clip(xa+w,ya+w,xb-w,yb-w);

  // Is there a string to write before the cursor ?
  if (string[stringstart]!='\0') {
    pen->Foreground((*set)[ocFOREGROUND]);
    pen->Write(x,y,string+stringstart);
    x+= pen->Width(string+stringstart);
  }
  // The cursor
  int z = m[0]!='\0' ? pen->Width(m) : pen->Width(" ");
  if (options & inFOCUS) {
    pen->Foreground((*set)[ocLIGHT]);
    pen->Fbox(x,ya+w+2,x+z,yb-w-2);
    pen->Foreground((*set)[ocFOREGROUND]);
  }
  pen->Write(x,y,m);
  x+=z;
  string[cursor] = m[0];

  // string after cursor
  if (string[cursor+1] != '\0') {
    if (options & inFOCUS) {
      pen->Foreground((*set)[ocFOREGROUND]);
    }
    pen->Write(x,y,string+cursor+1);
  }

  pen->Clip(0,0,0,0);
}

char INPUT::Press(uchar,int,int)
{
  if (!(options & inFOCUS)) window->Focus(this);
  return 0;
}

char INPUT::Motion(uchar,int,int)
{
  if (!(options & inFOCUS)) window->Focus(this);
  return 0;
}

char INPUT::Focus(FIGLET*)
{
  options |= inFOCUS;
  Refresh();  
  return 1;
}

void INPUT::DeFocus(FIGLET*)
{
  options &= 0x7F;
  Refresh();  
}

char INPUT::Input(char *m)
{
  char redraw = 0;
  ulong q = 0;
  while (m[q] != '\0') {
    switch (m[q]) {
    case '\r': // Pressing enter, showing the input away
    case '\n':
      if (cb_function) {
	if (cb_function(this,string)) {
	  String(NULL);
	  redraw = 1;
	}
      } else if (cb_element) {
	if (cb_element->Trigger(this)) {
	  String(NULL);
	  redraw = 1;
	}
      }
      break;
    case okDELETE: // deleting this character
      if (cursor<strlen(string)) {
	for (ulong i=cursor; string[i]!='\0'; i++) string[i] = string[i+1];
	redraw = 1;
      } else XBell(window->xdpy(),10);
      break;
    case okBACKSPACE: // deleting previous character
      if (cursor>0) {
	for (ulong i=cursor-1; string[i]!='\0'; i++) string[i] = string[i+1];
	cursor--; redraw = 1;
      } else XBell(window->xdpy(),10);
      break;

    case okHOME:
      cursor = 0;
      redraw = 1;
      break;
    case okEND:
      cursor = strlen(string);
      redraw = 1;
      break;
    case okLEFT:
      if (cursor>0) {
	cursor--;
	redraw = 1;
      } else XBell(window->xdpy(),10);
      break;
    case okRIGHT:
      if (cursor < strlen(string)) {
	cursor++;
	redraw = 1;      
      } else XBell(window->xdpy(),10);
      break;
    case okUP:
    case okDOWN:
    case okPAGEUP:
    case okPAGEDOWN:
      break;
    default: // add or overwrite character to existing string

      ulong p = strlen(string);
      if ((options & inOVERWRITE) && (cursor<p)) { // overwrite letter
	string[cursor] = m[q];
	cursor++;
	redraw = 1;
      } else {
	if (strlen(string)+5 >= stringsize) { // larger string by five
	  stringsize+=100;
	  char *s = new char[stringsize];
	  strcpy(s,string);
	  delete [] string;
	  string = s;
	}
	for (ulong i=p; i>cursor; i--)
	  string[i] = string[i-1];
	string[cursor] = m[q];
	cursor++;
	redraw = 1;
      }
      break;
    }
    q++;
  }
  if (redraw) DrawText();
  //printf("[%s]\n",string);
  return 1;
}
