
/*
   NAME:        otools.cc
   AUTHOR:      Ole Gunnar Westgaard TFU <olew@dina>
   DATE:        May 28 1996
   COMMENT:     otools base class functions
*/

#include "otools.h"

#include <unistd.h>
#include <sys/types.h>

#ifdef SGI
#include <bstring.h>
#endif // SGI

#include <sys/time.h>

#ifdef DEBUG_OTOOLS
 char D[128] = "";
 unsigned long Dl = D_ALL|D_DBUG|D_IBUGOBUG|D_HIGH|D_MED|D_LOW;
#endif

//=============================================================================
// OTOOLS
//=============================================================================

OTOOLS otools;

OTOOLS::OTOOLS(void)
  : displays(3,5)
{
#ifdef DEBUG_OTOOLS
  FILE *dbg = fopen("debug_value","rb");
  char line[1024];
  if (dbg) {
    Dl = 0;
    for (unsigned long i=0; i<32; i++) {
      if (fgets(line,1023,dbg) && (line[strlen(line)-2] == 'y'))
	Dl |= (1 << i);
    }
    fclose(dbg);
  }
#endif
  timer = NULL;
  fdread = NULL;
  fdwrite = NULL;
  fdexcept = NULL;
  running = 0;
}

char OTOOLS::Timer(ELEMENT *elem,ulong delay)
{
  IBUG(D_OTOOLS|D_MED,"OTOOLS[%p].Timer(ELEMENT[%p],delay=%ld)",
       this,elem,delay);
  TM *tm = timer;
  if (timer) {
    // stop if element allready exists... Only do update
    while (tm && (tm->element != elem) && tm->next) tm = timer->next;
    if (tm->element != elem) {
      tm->next = new TM;
      tm = tm->next;
    }
  } else {
    timer = new TM;
    tm = timer;
  }
  tm->next = NULL;
  tm->time = delay;
  tm->elapsed = 0;
  tm->element = elem;
  OBUG(D_OTOOLS|D_MED,"Timer set for element.");
  return 1;
}

char OTOOLS::UnTime(ELEMENT *elem)
{
  IBUG(D_OTOOLS|D_MED,"OTOOLS[%p].UnTime(ELEMENT[%p])",this,elem);
  if (!timer) return 0;
  TM *tm = timer,*prev = NULL;
  while (tm && (tm->element != elem)) {
    prev = tm;
    tm = tm->next;
  }
  if (tm) {
    if (prev)
      prev->next = tm->next;
    else
      timer = tm->next;
    tm->next = NULL;
    delete tm;
  }
  OBUG(D_OTOOLS|D_MED,"Timer removed for element.");
  return 1;
}

char OTOOLS::Add(DISPLAY *dpy)
{
  if (displays.append(dpy)) {
    DBUG(D_OTOOLS|D_LOW,"OTOOLS[%p].Add(DISPLAY[%p]) ok.",this,dpy);
    return 1;
  } else {
    DBUG(D_OTOOLS|D_LOW,"OTOOLS[%p].Add(DISPLAY[%p]) nack.",this,dpy);
    return 0;
  }
}

char OTOOLS::Del(DISPLAY *dpy)
{
  DBUG(D_OTOOLS|D_LOW,"OTOOLS[%p].Del(DISPLAY[%p])",this,dpy);
  if (displays.remove(dpy)) return 1; else return 0;
}

void OTOOLS::EventLoop(void)
{
  IBUG(D_OTOOLS|D_HIGH,"OTOOLS[%p].Eventloop()",this);
  TM *tm;
  DISPLAY *dpy;
  struct timeval tv;
  // XEvent event;
  fd_set set;
  running = 1;

  DBUG(D_OTOOLS|D_MED,"Flushing all displays.");
  for (ulong p = 1; (dpy = (DISPLAY*)displays[p]); ++p) {
    //dpy->Refresh();
    dpy->Flush();
  }

  DBUG(D_OTOOLS|D_MED,"Entering eventloop for all displays.");
  while ((displays.number_of_elements() > 0) && running) {
    int highest = -1;
    // INIT
    FD_ZERO(&set);
    // COLLECT DISPLAYS
    for (ulong p = 1; (dpy = (DISPLAY*)displays[p]); ++p) {
      //DBUG(D_OTOOLS|D_GARBAGE,"Collecting displays to monitor.");
      FD_SET(dpy->xfd(),&set);
      if (dpy->xfd() > highest) highest = dpy->xfd();
    }

    // WAIT FOR ACTIONS
    if (timer) {
      tv.tv_sec = 0;
      tv.tv_usec = 100;
      select(highest+1,&set,NULL,NULL,&tv);
    } else {
      DBUG(D_OTOOLS|D_GARBAGE,"Waiting for action.");
      select(highest+1,&set,NULL,NULL,NULL);
    }
      
    // TRIGGER ELEMENTS...
    if (timer) {
      DBUG(D_OTOOLS|D_GARBAGE,"Triggering timed elements.");
      tm = timer;
      while (tm) {
	tm->elapsed += 1;//tv.tv_sec*1000 + tv.tv_usec;
	if (tm->elapsed >= tm->time) {
	  TM *trig = tm;
	  tm = tm->next;
	  trig->elapsed = 0;
	  // it's allowed to remove the element from the queue
	  if (!trig->element->Trigger(this)) UnTime(trig->element);
	} else {
	  tm = tm->next;
	}
      }
    }
      
    // WHAT HAPPENED TO WHO ?
    for (ulong p = 1; (dpy = (DISPLAY*)displays[p]); ++p) {
      if (FD_ISSET(dpy->xfd(),&set)) {
	DBUG(D_GARBAGE,"Action at Display -- %p",dpy);
	XProcessInternalConnection(dpy->xdpy(),dpy->xfd());
	dpy->HandleCurrentEvents();
	//XNextEvent(dpy->xdpy(),&event);
	//dpy->HandleSingleEvent(&event);
	dpy->Flush();
      }
    }

  }
  FD_ZERO(&set);
  OBUG(D_OTOOLS|D_HIGH,"Eventloop done.");
}


void OTOOLS::Stop(void)
{
  DBUG(D_OTOOLS|D_MED,"OTOOLS[%p].Stop()",this);
  running = 0;
}

//=============================================================================
// DISPLAY
//=============================================================================

DISPLAY::DISPLAY(char *display_name)
  : colormaps(5,5),
    windows(10,10)
{
  // Open X11 display, display will be NULL if it fails
  IBUG(D_DISPLAY|D_HIGH,"new DISPLAY[%p](\"%s\")",this,display_name);
  if (!(display = XOpenDisplay(display_name))) {
    WARN("new DISPLAY[%p](\"%s\") cannot open display",this,display_name);
    OBUG(D_DISPLAY|D_HIGH,"Failed.");
    return;
  }
  {
    char *n = XDisplayName(display_name);
    title = new char[strlen(n)+1];
    strcpy(title,n);
  }
  screen  = DefaultScreen(display);
  DBUG(D_DISPLAY|D_MED,"Display opened on screen %d",screen);
  width   = DisplayWidth(display,screen);
  height  = DisplayHeight(display,screen);
  DBUG(D_DISPLAY|D_MED,"Displaysize = (%d,%d)",width,height);
  running = 0;
  // Create the DEFAULT colormap, it is NOT part of the COLORMAP list
  defcmap = new COLORMAP(this,DefaultColormap(display,screen));
  otools.Add(this);
  XSynchronize(display,False);
  OBUG(D_DISPLAY|D_HIGH,"Ok.");
}

DISPLAY::~DISPLAY(void)
{
  WINDOW *w;
  COLORMAP *c;
  IBUG(D_DISPLAY|D_HIGH,"~DISPLAY[%p]",this);
  while ((w = (WINDOW*)windows[1])) {
    DBUG(D_DISPLAY|D_MED,"Removing 'forgotten' window[%p]",w);
    delete w; // Will remove itself from the list
  }
  delete defcmap;
  while ((c = (COLORMAP*)colormaps[1])) {
    DBUG(D_DISPLAY|D_MED,"Removing 'forgotten' colormap[%p]",c);
    delete c; // Will remove itself from the list
  }
  XCloseDisplay(display);
  otools.Del(this);
  if (title) delete [] title;
  OBUG(D_DISPLAY|D_MED,"Deleted.");
}

void DISPLAY::HandleCurrentEvents(void)
{
  IBUG(D_DISPLAY|D_GARBAGE,"DISPLAY[%p].HandleCurrentEvents()",this);
  while (XPending(display)) HandleSingleEvent();
  OBUG(D_DISPLAY|D_GARBAGE,"Done.");
}

void DISPLAY::EventLoop(void)
{
  IBUG(D_DISPLAY|D_MED,"DISPLAY[%p].EventLoop()",this);
  XEvent event;
  running = 1;
  while (running) {
    // Wait for an event
    XNextEvent(display,&event);
    HandleSingleEvent(&event);
    if (windows.number_of_elements() == 0) running = 0;
  }
  OBUG(D_DISPLAY|D_MED,"Done.");
}

void DISPLAY::HandleSingleEvent(XEvent *event)
  /// Handles inactive windows, use when an event has been collected.
{
  IBUG(D_DISPLAY|D_GARBAGE,"DISPLAY[%p].HandleSingleEvent(%d)",
       this,event->type);
  WINDOW *w;
  if ((w = (WINDOW*)windows.find((ulong)event->xany.window)))
    w->HandleEvent(event);
  OBUG(D_DISPLAY|D_GARBAGE,"Done.");
}

void DISPLAY::HandleSingleEvent(void)
  /// Assumes you don't know if there's an event there, will wait for one.
  /// Does not check wether a window is disabled or not.
{
  IBUG(D_DISPLAY|D_GARBAGE,"DISPLAY[%p].HandleSingleEvent()",this);
  WINDOW *w;
  XEvent event;
  // Wait for an event
  XNextEvent(display,&event);
  // Handle it
  if ((w = (WINDOW*)windows.find((ulong)event.xany.window))) {
    //DBUG(D_DISPLAY|D_GARBAGE,"Handling event for window -- %p",w);
    w->HandleEvent(&event);
  }
  OBUG(D_DISPLAY|D_GARBAGE,"Done.");
}

void DISPLAY::NoWaitEvent(void)
  /// Checks once for each window, can throw cpy usage to 1.0 for the process.
{
  WINDOW *w;
  XEvent event;
  for (ulong i = 1; (w = (WINDOW*)windows[i]); ++i)
    if (XCheckWindowEvent(display,w->xwnd(),0xFFFFFFFF,&event))
      w->HandleEvent(&event);
}

//=============================================================================
// COLORMAP
//=============================================================================

COLORMAP::COLORMAP(DISPLAY *dpy,Colormap map)
  : colorsets(10,10)
{
  // Create a COLORMAP using the allready created Colormap
  // Don't add this colormap to the display
  display = dpy;
  colormap = map;
}

COLORMAP::COLORMAP(DISPLAY *dpy)
  : colorsets(10,10)
{
  // Create a new COLORMAP using display
  display = dpy;
  colormap = XCreateColormap(dpy->xdpy(),dpy->xroot(),
			     DefaultVisual(dpy->xdpy(),dpy->xscr()),
			     AllocNone);
  if (!colormap) {
    WARN("COLORMAP() Could not create X11 Colormap!%c%c%c",0,0,0);
    colormap = 0;
    display = NULL;
    return;
  }
  display->Add(this);
}

COLORMAP::~COLORMAP(void)
{
  IBUG(D_COLORMAP|D_HIGH,"~COLORMAP[%p]()",this);
  if (!display) {
    WARN("~COLORMAP[%p]() Removing non-initialized colormap!",this);
    OBUG(D_COLORMAP|D_HIGH,"Not initialized.");
    return;
  }
  if (colorsets.number_of_elements()) {
    // Colorsets have not been removed
    COLORSET *c;
    DBUG(D_COLORMAP|D_MED,"Removing %ld 'forgotten' colorsets from colormap!",
	 colorsets.number_of_elements());
    while ((c = (COLORSET*)colorsets[1])) {
      colorsets.remove(c);
      delete c;
    }
  }
  if (display->Del(this))
    XFreeColormap(display->xdpy(),colormap);
  OBUG(D_COLORMAP|D_HIGH,"Deleted.");
}

ushort COLORMAP::Value(char *s)
{
  // "a" "aa" "aaa" "aaaa"
  // Converts a 1-4 digit hexadecimal to unsigned short value
  // Only digits as parameter,
  ushort v=0;
  int p=0;
  while (s[p] != '\0') {
    v = v << 4;
    if ((s[p]>='0') && (s[p]<='9')) v += (s[p]-'0'); else
    if ((s[p]>='a') && (s[p]<='f')) v += (s[p]-'a'+10); else
    if ((s[p]>='A') && (s[p]<='F')) v += (s[p]-'A'+10);
    ++p;
  }
  return v;
}

char COLORMAP::AllocateColor(char *name,ulong *index)
{
  // This function will only allocate ONE color!
  XColor cx,cc;
  if (name[0] == '#') {
    char tmp;
    if ((strlen(name)-1) == 3) { // 4 bits per color
      tmp = name[2]; name[2] = '\0';
      cx.red = Value(name+1) * 0x1111;
      name[2] = tmp; tmp = name[3]; name[3] = '\0';
      cx.green = Value(name+2) * 0x1111;
      name[3] = tmp;
      cx.blue = Value(name+3) * 0x1111;
    } else
    if ((strlen(name)-1) == 6) { // 8 bits per color
      tmp = name[3]; name[3] = '\0';
      cx.red = Value(name+1) * 0x101;
      name[3] = tmp; tmp = name[5]; name[5] = '\0';
      cx.green = Value(name+3) * 0x101;
      name[5] = tmp;
      cx.blue = Value(name+5) * 0x101;
    } else
    if ((strlen(name)-1) == 9) { // 12 bits per color
      tmp = name[4]; name[4] = '\0';
      cx.red = Value(name+1) * 0x10;
      name[4] = tmp; tmp = name[7]; name[7] = '\0';
      cx.green = Value(name+4) * 0x10;
      name[7] = tmp;
      cx.blue = Value(name+7) * 0x10;
    } else
    if ((strlen(name)-1) == 12) { // 16 bits per color
      tmp = name[5]; name[5] = '\0';
      cx.red = Value(name+1);
      name[5] = tmp; tmp = name[9]; name[9] = '\0';
      cx.green = Value(name+5);
      name[9] = tmp;
      cx.blue = Value(name+9);
    } else {
      WARN("COLORMAP[%p].AllocateColor(\"%s\") Wrong hex format color!",
	   this,name);
      return 0;
    }
    if (XAllocColor(display->xdpy(),colormap,&cx)) {
      *index = cx.pixel;
      // DBUG("COLORMAP.GetColor('%s') allocated, index=%lX%c",name,*index,0);
      return 1;
    }
  } else {
    if (XAllocNamedColor(display->xdpy(),colormap,name,&cc,&cx)) {
      *index = cc.pixel;
      // DBUG("COLORMAP.GetColor('%s') allocated, index=%lX%c",name,*index,0);
      return 1;
    }
  }
  WARN("COLORMAP[%p].AllocateColor(\"%s\") cannot allocate color!",
       this,name);  
  return 0;
}

void COLORMAP::FreeColors(ulong *index,int num)
{
  XFreeColors(display->xdpy(),colormap,index,num,0);
  // DBUG("COLORMAP.FreeColors(index,%d) ok%c%c",num,0,0);
}

//=============================================================================
// COLORSET
//=============================================================================

COLORSET::COLORSET(COLORMAP *cmap,char *cnames)
  /// Creates a colorset on COLORMAP from the colon separated string.
  /// Commas at the end of the list are illegal!!
{
  IBUG(D_COLORSET|D_HIGH,"new COLORSET[%p](COLORMAP[%p],\"%s\")",
       this,cmap,cnames);
  colormap = cmap;
  count = 0;
  set = NULL;
  colors = NULL;
  name_mem = NULL;
  name_mem_size = 0;
  if (Add(cnames) == -1) {
    colormap = NULL;
    OBUG(D_COLORSET|D_HIGH,"Failed!");
    return;
  }
  colormap->Add(this);                               // Enroll me with colormap
  OBUG(D_COLORSET|D_HIGH,"Done.");
}

COLORSET::~COLORSET(void)
{
  IBUG(D_COLORSET|D_HIGH,"~COLORSET[%p]()",this);
  if (!colormap) {
    WARN("~COLORSET[%p]() was not initialised correctly!",this);
    OBUG(D_COLORSET|D_HIGH,"Not initialized.");
    return;
  }
  colormap->FreeColors(set,count);                           // Free all colors
  delete [] set;                                        // Remove the index set
  delete [] colors;                                 // Delete color table names
  delete [] name_mem;                                     // Delete name memory
  colormap->Del(this);            // Don't want to be in colormaps list anymore
  OBUG(D_COLORSET|D_HIGH,"Deleted.");
}

int COLORSET::Add(char *newco)
  /// Adds colors, updates count, returns index to first in list
  /// Returns negative number if it fails
  /// -1 cannot copy string to internal memory
  /// -2 cannot allocate memory for nameindex
  /// -3 no colors to allocate
  /// -4 could not allocate all colors
  /// -5 cannot allocate memory to hold all colors
{

  DBUG(D_COLORSET|D_LOW,"COLORS[%p].Add(\"%s\") count=%d",this,newco,count);

  // COPY STRING OF COLORS
  int tnms = strlen(newco)+1;                     // temporary name memory size
  char *cs;                                        // contains string of colors
  if (!(cs = new char[tnms])) {                              // cannot allocate
    WARN("COLORS[%p].Add(\"%s\") Cannot allocate memory (1)!",this,newco);
    return -1;
  }
  strcpy(cs,newco);

  DBUG(D_COLORSET|D_LOW,"tnms=%d cs=\"%s\"",tnms,cs);

  // SPLIT INTO COLORNAMES WITH STRTOK
  int    ni=0,mi=20;                                    // name-index,max-index
  char **tni;
  if (!(tni = new (char*)[mi])) {                            // cannot allocate
    WARN("COLORS[%p].Add(\"%s\") Cannot allocate memory (2)!",this,newco);
    delete [] cs;
    return -2;
  }
  tni[ni] = strtok(cs,":;,./\\");                                // first timer
  while (tni[ni]) {                                // split into separate names
    ++ni;
    if (ni >= mi) {
      char **tmp;
      mi += 20;
      if ((tmp = new (char*)[mi])) {          // allocate more nameindex memory
	for (int i=0; i<=ni; i++) tmp[i] = tni[i];
	delete [] tni;
	tni = tmp;
      } else {                                               // cannot allocate
	delete [] cs;
	delete [] tni;
	return -2;
      }
    }
    tni[ni] = strtok(NULL,":;,./\\");                              // next name
  }                                        // ni = number of colors to allocate

  DBUG(D_COLORSET|D_LOW,"ni=%d mi=%d tni=%p",ni,mi,tni);

  // CHECK THAT THERE ARE COLORS TO ALLOCATE
  if (ni==0) {                                         // no colors to allocate
    WARN("COLORS[%p].Add(\"%s\") No colors to allocate!",this,newco);
    delete [] cs;
    delete [] tni;
    return -3;
  }

  // ALLOCATE THE NEW COLORS
  ulong tci[ni];                           // We need exactly this many indexes
  for (int i=0; i<ni; i++) {                             // allocate each color
    if (!colormap->AllocateColor(tni[i],&tci[i])) {    // allocate single color
      WARN("COLORS[%p].Add(\"%s\") Could not allocate color!",this,newco);
      if (i) colormap->FreeColors(tci,i);                        // free colors
      delete [] cs;
      delete [] tni;
      return -4;                               // could not allocate all colors
    }
  }

  // ALLOCATE NEW MEMORY FOR ALL COLORS IN COLORSET
  ulong *nci;                            // new color indexes, both old and new
  char  *nnm;                                                // new name memory
  char  **nni;                                              // new name indexes
  if (!(nci = new ulong[ni+count])) {             // allocate new color indexes
    WARN("COLORS[%p].Add(\"%s\") Could not allocate memory (1)!",this,newco);
    colormap->FreeColors(tci,ni);
    delete [] cs;
    delete [] tni;
    return -5;
  }
  if (!(nnm = new char[tnms+name_mem_size])) {      // allocate new name memory
    WARN("COLORS[%p].Add(\"%s\") Could not allocate memory (2)!",this,newco);
    colormap->FreeColors(tci,ni);
    delete [] cs;
    delete [] tni;
    delete [] nci;
    return -5;
  }
  if (!(nni = new (char*)[ni+count])) {            // allocate new name indexes
    WARN("COLORS[%p].Add(\"%s\") Could not allocate memory (3)!",this,newco);
    colormap->FreeColors(tci,ni);
    delete [] cs;
    delete [] tni;
    delete [] nci;
    delete [] nnm;
    return -5;
  }

  // COPY OLD COLORS INTO NEW MEMORY
  if (set) {
    memcpy(nnm,name_mem,name_mem_size);                  // memcopy name memory
    for (int i=0; i<count; i++) {
      nci[i] = set[i];                                    // Copy color indexes
      nni[i] = nnm + (colors[i] - name_mem);    // Translate color name indexes
    }
  }  

  // COPY NEW COLORS INTO NEW MEMORY
  for (int i=count; i<(count+ni); i++) {
    nci[i] = tci[i-count];                                // copy color indexes
    nni[i] = (nnm+name_mem_size) + (tni[i-count]-cs);      // copy name indexes
  }
  memcpy(nnm+name_mem_size,cs,tnms);    // memcopy (cat) colorstring to new mem

  // REMOVE OLD COLORMEMORY
  if (set) {
    delete [] name_mem;
    delete [] colors;
    delete [] set;
  }

  // SET MEMORY TO POINT TO NEW MEMORY
  int rc = count;
  count += ni;
  set = nci;
  colors = nni;
  name_mem = nnm;
  name_mem_size += tnms;
 
  // RETURN INDEX TO NEW COLORS
  //for (int i=0; i<name_mem_size; i++)
  //  if (name_mem[i] == '\0')
  //    ::printf(":");
  //  else
  //    ::printf("%c",name_mem[i]);
  //::printf("\n");

  return rc;
}

char COLORSET::Change(int p,char *name)
{

  // ALLOCATE NEW COLOR
  if (!colormap || !Exist(p)) return 0;     // Must be initialized and an index

  ulong c;
  {
    char *myname = new char[strlen(name)+1];
    strcpy(myname,name);
    if (!colormap->AllocateColor(myname,&c)) return 0;        // Allocate success ?
    delete [] myname;
  }
    
   // FIX NAME MEMORY AND INDEXES
  if (strlen(name) == strlen(colors[p])) {          // if the name is same size
    colormap->FreeColors(&set[p],1);                        // Remove old color
    set[p] = c;                                                // Use new color
    strcpy(colors[p],name);
    return 1;
  }
 
  int diff = (strlen(name)-strlen(colors[p]));
  int nmsize = name_mem_size + diff;
  char *nnm;

  // CREATE MEMORY
  if (!(nnm = new char[nmsize])) {
    colormap->FreeColors(&c,1);
    return 0;
  }
 
  memcpy(nnm,name_mem,(colors[p]-name_mem));
  if (p<count-1)
    memcpy(nnm+(colors[p]-name_mem)+strlen(name)+1,
	   colors[p+1],
	   (name_mem_size - (colors[p+1]-name_mem)));
  for (int i=0; i<p+1; i++)
    colors[i] = nnm+(colors[i]-name_mem);
  for (int i=p+1; i<count; i++)
    colors[i] = nnm+(colors[i]-name_mem)+diff;
  delete [] name_mem;
  name_mem_size = nmsize;
  name_mem = nnm;
  strcpy(colors[p],name);
  colormap->FreeColors(&set[p],1);                          // Remove old color
  set[p] = c;                                                  // Use new color
 
  //for (int i=0; i<name_mem_size; i++)
  //  if (name_mem[i] == '\0')
  //    ::printf(":");
  //  else
  //    ::printf("%c",name_mem[i]);
  //::printf("\n");

  return 1;
}

//=============================================================================
// WINDOW
//=============================================================================
int xerrorhandler(Display*,XErrorEvent*)
{
  printf("SERIOUS X ERROR\n");
  return 1;
}

char WINDOW::Create(DISPLAY *_dp,COLORMAP *_mp,COLORSET *_cs,Window _pt,
		    char *_tt,char *_it,int _xp,int _yp,int miw,int mih,
		    int inw,int inh,int maw,int mah,int _bw,uchar *_ic,
		    uint _iw,uint _ih,char *_fn,int _dt,Visual *_vs,uchar _mt)
{
  // Parent must exist on the created windows display, a value of 0 will use
  // this setting "RootWindow(display->xdpy(),display->xscr())"

  // INITIAL VALUES ===========================================================

  //XSetErrorHandler(xerrorhandler);

  XSetWindowAttributes xwa;
  XTextProperty        wName,
                       iName;
  XSizeHints           shints;
  XClassHint           chints;
  XWMHints             wmhints;
  ulong                xwamap;

  IBUG(D_WINDOW|D_MED,"WINDOW[%p].Create(...,\"%s\",...)",this,_tt);
  
  // CHECKING PARAMETERS ======================================================

  if (!_cs) {
    WARN("WINDOW[%p].Create(...,\"%s\",...) Need colorset to create windows!",
	 this,_tt);
    display = NULL;
    return 0;
  }

  colorset = _cs;
  colormap = _mp ? _mp : colorset->cmap();
  display  = _dp ? _dp : colormap->dpy();
  focused  = NULL;
  selected = NULL;
  bd_idx   = ocBORDER;
  bg_idx   = ocBACKGROUND;
  fg_idx   = ocFOREGROUND;

  if (!_pt) _pt = RootWindow(display->xdpy(),display->xscr()); // same display!
  if (!_dt) _dt = CopyFromParent;
  if (!_vs) _vs = CopyFromParent;

  // SETTING XWINDOW ATTRIBUTES ===============================================
  DBUG(D_WINDOW|D_LOW,"Setting window attributes.");

  //xwa.background_pixmap     = None;
  //xwa.border_pixmap         = CopyFromParent;
  //xwa.bit_gravity           = ?;
  //xwa.win_gravity           = ?;
  //xwa.backing_store         = WhenMapped; // NotUseful,WhenMapped,Always
  //xwa.backingplanes         = ?;
  //xwa.backingpixel          = ?;
  //xwa.save_under            = true | false;
  //xwa.cursor                = ?;

  xwa.background_pixel      = (*colorset)[ocBACKGROUND];
  xwa.border_pixel          = (*colorset)[ocBORDER];
  xwa.event_mask            = ExposureMask      | KeyPressMask        |
                              KeyReleaseMask    | ButtonPressMask     |
                              ButtonReleaseMask | StructureNotifyMask |
                              PointerMotionMask | EnterWindowMask     |
                              LeaveWindowMask   | OwnerGrabButtonMask ;
  xwa.do_not_propagate_mask = None;
  xwa.override_redirect     = true;
  xwa.colormap              = colormap->xcmap();

  // CWBackPixmap    | CWBackPixel        | CWBorderPixmap | CWBorderPixel
  // CWBitGravity    | CWWinGravity       | CWBackingStore | CWBackingPlanes
  // CWBackingPixel  | CWOverrideRedirect | CWSaveUnder    | CWEventMask
  // CWDontPropagate | CWColormap         | CWCursor

  xwamap = CWBackPixel | CWBorderPixel | CWEventMask | CWDontPropagate |
           CWColormap  ;
  if (_mt & WINDOWISMENU) xwamap |= CWOverrideRedirect;


  // CREATING WINDOW ==========================================================
  DBUG(D_WINDOW|D_LOW,"Creating window.");

  window = XCreateWindow(display->xdpy(),_pt,_xp,_yp,inw,inh,_bw,
			 _dt,InputOutput,_vs,xwamap,&xwa);
  if (!window) {
    WARN("WINDOW[%p].Create(...,\"%s\",...) Cannot create window!",
	 this,_tt);
    display = NULL;
    return 0;
  }

  posx   = _xp;
  posy   = _yp;
  width  = inw;
  height = inh;

  // CREATING ICON ============================================================


  if (!_ic) icon = 0; else {

    DBUG(D_WINDOW|D_LOW,"Creating icon.");

    icon = XCreateBitmapFromData(display->xdpy(),window,
				 (const char*)_ic,_iw,_ih);
    if (!icon) {
      WARN("WINDOW[%p].Create(...,\"%s\",...) Cannot create icon!",
	   this,_tt);
      XDestroyWindow(display->xdpy(),window);
      display = NULL;
      return 0;
    }
  }

  // WINDOW MANAGER OPTIONS ===================================================
  
  DBUG(D_WINDOW|D_LOW,"Window manager options.");

  title = new char[strlen(_tt)+1];
  strcpy(title,_tt);
  XStringListToTextProperty(&_tt,1,&wName);
  XStringListToTextProperty(&_it,1,&iName);
  shints.flags          = PPosition|PSize|PMinSize|PMaxSize;
  shints.min_width      = miw;
  shints.min_height     = mih;
  shints.max_width      = maw;
  shints.max_height     = mah;
  wmhints.initial_state = NormalState;
  wmhints.input         = True;
  wmhints.icon_pixmap   = icon;
  wmhints.flags         = StateHint | IconPixmapHint | InputHint ;
  chints.res_name       = title;
  chints.res_class      = "otools";
  XSetWMProperties(display->xdpy(),window,&wName,&iName,NULL,0,    // argv,argc
		   &shints,&wmhints,&chints);

  // PEN AND COLORMAPS ========================================================

  DBUG(D_WINDOW|D_LOW,"Creating pen.");

  pen = new PEN(this,_fn);
  if (!pen->wnd()) {
    WARN("WINDOW[%p].Create(...,\"%s\",...) Cannot create default pen!",
	 this,_tt);
    delete pen;
    XDestroyWindow(display->xdpy(),window);
    if (icon) XFreePixmap(display->xdpy(),icon);
    display = NULL;
    return 0;
  }

  // FINISHING OFF THE CREATION ===============================================

  OBUG(D_WINDOW|D_MED,"Done.");
  return 1;
}

WINDOW::WINDOW(void)
  : display(NULL),
    refresh(10,10),
    active(10,10),
    motion(10,10),
    pens(5,10)
{
  DBUG(D_WINDOW|D_LOW,"new WINDOW[%p]() Null creator.",this);
}

WINDOW::WINDOW(COLORSET *set,char *tt,char *fnt,int x,int y,int w,int h)
  : display(NULL),
    refresh(10,10),
    active(10,10),
    motion(10,10),
    pens(5,10)
{
  IBUG(D_WINDOW|D_MED,"new WINDOW[%p](\"%s\") minimal creator",this,tt);
  if (Create(set->cmap()->dpy(),set->cmap(),set,0,tt,tt,x,y,w,h,w,h,w,h,
	     1,NULL,0,0,fnt,0,NULL,0)) {
    DBUG(D_WINDOW|D_LOW,"Mapping and registering window to display '%s'",
	 display->name());
    display->Add(this);
    XMapWindow(display->xdpy(),window);
    OBUG(D_WINDOW|D_MED,"Created.");
  } else {
    WARN("new WINDOW[%p](minimal creator) Cannot create window!",this);
    OBUG(D_WINDOW|D_MED,"Not Created!");
  }
}

WINDOW::WINDOW(COLORSET *set,char *tt,char *fnt,int x,int y,int w,int h,
	       uchar *icn,uint iw,uint ih)
  : display(NULL),
    refresh(10,10),
    active(10,10),
    motion(10,10),
    pens(5,10)
{
  IBUG(D_WINDOW|D_MED,"new WINDOW[%p](\"%s\") medium creator.",this,tt);
  if (Create(set->cmap()->dpy(),set->cmap(),set,0,tt,tt,x,y,w,h,w,h,w,h,
	     1,icn,iw,ih,fnt,0,NULL,0)) {
    DBUG(D_WINDOW|D_LOW,"Mapping and registering window to display '%s'",
	 display->name());
    display->Add(this);                               // Enroll me with display
    XMapWindow(display->xdpy(),window);
    OBUG(D_WINDOW|D_MED,"Created.");
  } else {
    WARN("new WINDOW[%p](medium creator) Cannot create window!",this);
    OBUG(D_WINDOW|D_MED,"Not Created!");
  }
}

WINDOW::WINDOW(uchar mtd,WINDOW *pnt,COLORSET *set,char *tt,char *it,int x,
	       int y,int miw,int mih,int inw,int inh,int maw,int mah,int bdw,
	       uchar *icn,uint icw,uint ich,char *fnt)
  : display(NULL),
    refresh(10,10),
    active(10,10),
    motion(10,10),
    pens(10,5)
{
  IBUG(D_WINDOW|D_MED,"new WINDOW[%p](\"%s\") maximal creator",this,tt);
  if (Create(set->cmap()->dpy(),set->cmap(),set,
	     pnt ? pnt->xwnd() : 0,tt,it,x,y,miw,mih,
	     inw,inh,maw,mah,bdw,icn,icw,ich,fnt,0,NULL,mtd)) {
    DBUG(D_WINDOW|D_LOW,"Mapping and registering window to display '%s'",
	 display->name());
    display->Add(this);                               // Enroll me with display
    XMapWindow(display->xdpy(),window);
    OBUG(D_WINDOW|D_MED,"Created.");
  } else {
    WARN("new WINDOW[%p](maximal creator) Cannot create window!",this);
    OBUG(D_WINDOW|D_MED,"Not Created!");
  }
}

WINDOW::~WINDOW(void)
{
  IBUG(D_WINDOW|D_MED,"~WINDOW[%p]()",this);
  PEN *pe;
  FIGLET *f;
  if (!display) {                             // Check if initialization worked
    OBUG(D_WINDOW|D_MED,"Not initialized.");
    return;
  }
  display->Del(this);                 // Don't want to be part of displays list
  while ((f = (FIGLET*)refresh[1])) {
    refresh.remove(f);
    delete f;
  }
  while ((pe = (PEN*)pens[1])) {
    pens.remove(pe);
    delete pe;
  }
  if (icon)
    XFreePixmap(display->xdpy(),icon);          // Don't need this icon no more
  XUnmapWindow(display->xdpy(),window);                     // Unmapping window
  XDestroyWindow(display->xdpy(),window);                  // Remove the window
  if (title) delete [] title;
  OBUG(D_WINDOW|D_MED,"Deleted.");
}

void WINDOW::HandleEvent(XEvent *event)                // Handle incoming event
{
  static int x=0,y=0;
  static uchar b=0;
  static ulong i=0;
  static FIGLET *fig      = NULL,
                *found    = NULL,
                *oldfig   = NULL;

  switch (event->type) {

  //---------------------------------------------------------------------------
  case MapNotify:
    DBUG(D_WINDOW|D_LOW,"WINDOW[%p].MapNotify",this);
    break;

  //---------------------------------------------------------------------------
  case ReparentNotify:
    DBUG(D_WINDOW|D_LOW,"WINDOW[%p].ReparentNotify",this);
    break;

  case DestroyNotify:
    DBUG(D_WINDOW|D_LOW,"WINDOW[%p].DestroyNotify",this);
    break;

  case UnmapNotify:
    DBUG(D_WINDOW|D_LOW,"WINDOW[%p].UnmapNotify",this);
    break;

  case NoExpose:
    // Appears in cooperation with XCopyArea... XCopyPlane
    DBUG(D_WINDOW|D_LOW,"WINDOW[%p].NoExpose",this);
    break;

  //===========================================================================
  //= EXPOSE =============================== exposing whole or part of window =
  //===========================================================================

  case Expose:
    IBUG(D_WINDOW|D_LOW,"WINDOW[%p].Expose(%d,%d,%d,%d)",this,
	 event->xexpose.x,event->xexpose.y,
	 event->xexpose.width,event->xexpose.height);
    Before(event->xexpose.x,event->xexpose.y,
	   event->xexpose.width,event->xexpose.height);
    Refresh(event->xexpose.x,event->xexpose.y,
            event->xexpose.width,event->xexpose.height);
    After(event->xexpose.x,event->xexpose.y,
	  event->xexpose.width,event->xexpose.height);
    OBUG(D_WINDOW|D_LOW,"Done expose.");
    break;

  //===========================================================================
  //= CONFIGURENOTIFY ======================= resizing and position of window =
  //===========================================================================

  case ConfigureNotify: // may appear _after_ expose (!)
    IBUG(D_WINDOW|D_LOW,"WINDOW[%p].ConfigureNotify(%d,%d,%d,%d)",this,
	 event->xconfigure.x,event->xconfigure.y,
	 event->xconfigure.width,event->xconfigure.height);
    if ((event->xconfigure.x != 0) || (event->xconfigure.y != 0)) {
      posx = event->xconfigure.x;
      posy = event->xconfigure.y;
      Move();
    }
    if ((width != (event->xconfigure.width -1)) ||
	(height != (event->xconfigure.height-1))) {
      width = event->xconfigure.width;        // Get new width and height
      height = event->xconfigure.height;
      Size();
      Clear();    // Clear window (Expose will follow after configure-resize)
      Refresh(0,0,width,height);
    }                               // We don't care about window positioning
    OBUG(D_WINDOW|D_LOW,"Done configure.");
    break;
    
  //===========================================================================
  //= BUTTONPRESS ======================================= pressing any button =
  //===========================================================================

  case ButtonPress:
    IBUG(D_WINDOW|D_LOW,"WINDOW[%p].ButtonPress(%d,%d,%d)",this,
	 event->xbutton.button,event->xbutton.x-1,event->xbutton.y-1);
    b = event->xbutton.button;
    x = event->xbutton.x-1;
    y = event->xbutton.y-1;
    // If selected, send motion to it, deselect it if it wants to
    oldfig = selected;
    if (selected && (!selected->Press(b,x,y))) selected = NULL;
    // If noone is selected, check if there's another figlet around
    if (!selected) {
      found = NULL;
      for (i=1; (!found) && (fig = (FIGLET*)active[i]); ++i)
	if (fig->Inside(x,y)) found = fig;
      // send press to figlet if found and it's not oldfig
      if (found && (oldfig!=found) && found->Press(b,x,y)) selected = found;
    }
    OBUG(D_WINDOW|D_LOW,"Done ButtonPress.");
    break;

  //===========================================================================
  //= BUTTONRELEASE ==================================== releasing any button =
  //===========================================================================

  case ButtonRelease:
    IBUG(D_WINDOW|D_LOW,"WINDOW[%p].ButtonRelease(%d,%d,%d)",this,
	 event->xbutton.button,event->xbutton.x-1,event->xbutton.y-1);
    b = event->xbutton.button;
    x = event->xbutton.x-1;
    y = event->xbutton.y-1;
    // If selected, send motion to it, deselect it if it wants to
    oldfig = selected;
    if (selected && (!selected->Release(b,x,y))) selected = NULL;
    // If noone is selected, check if there's another figlet around
    if (!selected) {
      found = NULL;
      for (i=1; (!found) && (fig = (FIGLET*)active[i]); ++i)
	if (fig->Inside(x,y)) found = fig;
      // send press to figlet if found and it's not oldfig
      if (found && (oldfig!=found) && found->Release(b,x,y)) selected = found;
    }
    OBUG(D_WINDOW|D_LOW,"Done ButtonRelease.");
    break;

  //===========================================================================
  //= MOTIONNOTIFY ================================== when mousepointer moves =
  //===========================================================================

  case MotionNotify:
    IBUG(D_WINDOW|D_LOW,"WINDOW[%p].MotionNotify(%d,%d,%d)",this,
	 event->xmotion.state,event->xbutton.x-1,event->xbutton.y-1);
    b = event->xmotion.state>>8;
    x = event->xbutton.x-1;
    y = event->xbutton.y-1;
    // If selected, send motion to it, deselect it if it wants to
    oldfig = selected;
    if (selected && (!selected->Motion(b,x,y))) selected = NULL;
    // check if we're inside a motion-figlet if no figlet is selected
    if (!selected) {
      found = NULL;
      for (i=1; (!found) && (fig = (FIGLET*)motion[i]); ++i)
	if (fig->Inside(x,y)) found = fig;
      // send motion if a motion figlet is found and it's not oldfig
      if (found && (oldfig!=found) && found->Motion(b,x,y)) selected = found;
    }
    OBUG(D_WINDOW|D_LOW,"Done MotionNotify.");
    break;

  //===========================================================================
  //= ENTERNOTIFY =========================== when mousepointer enters window =
  //===========================================================================

  case EnterNotify:
    IBUG(D_WINDOW|D_LOW,"WINDOW[%p].EnterNotify(%d,%d)",this,
	 event->xbutton.x-1,event->xbutton.y-1);
    Enter(event->xbutton.x-1,event->xbutton.y-1);
    OBUG(D_WINDOW|D_LOW,"Done EnterNotify.");
    break;

  //===========================================================================
  //= LEAVENOTIFY =========================== when mousepointer leaves window =
  //===========================================================================

  case LeaveNotify:
    IBUG(D_WINDOW|D_LOW,"WINDOW[%p].LeaveNotify(%d,%d)",this,
	 event->xbutton.x-1,event->xbutton.y-1);
    Leave(event->xbutton.x-1,event->xbutton.y-1);
    OBUG(D_WINDOW|D_LOW,"Done LeaveNotify.");
    break;
    
  //===========================================================================
  //= KEYPRESS ============================== when a key is pressed in window =
  //===========================================================================

  case KeyPress:
    {
      KeySym keysym;
      XComposeStatus compose;
      char buffer[40];
      int count = XLookupString((XKeyEvent*)event,buffer,39,&keysym,&compose);
      buffer[count]='\0';
      DBUG(D_WINDOW|D_LOW,"WINDOW[%p].KeyPress(\"%s\")",this,buffer);
      if (focused) {
	if ((keysym >= XK_KP_Space) && (keysym<=XK_KP_9))
	  focused->Input(buffer);
	else if ((keysym >= XK_space) && (keysym <= XK_asciitilde))
	  focused->Input(buffer);
	else if ((keysym >= XK_nobreakspace) && (keysym <= XK_ydiaeresis))
	  focused->Input(buffer);
	else switch (keysym) {
	case XK_Return: case XK_KP_Enter: case XK_Linefeed: case XK_Delete:
	case XK_BackSpace:
	  focused->Input(buffer);
	  break;
	case XK_Left: case XK_KP_Left:
	  buffer[count++] = okLEFT;
	  buffer[count] = '\0';
	  focused->Input(buffer);
	  break;
	case XK_Right: case XK_KP_Right:
	  buffer[count++] = okRIGHT;
	  buffer[count] = '\0';
	  focused->Input(buffer);
	  break;
	case XK_Up: case XK_KP_Up:
	  buffer[count++] = okUP;
	  buffer[count] = '\0';
	  focused->Input(buffer);
	  break;
	case XK_Down: case XK_KP_Down:
	  buffer[count++] = okDOWN;
	  buffer[count] = '\0';
	  focused->Input(buffer);
	  break;
	case XK_Home: case XK_KP_Home:
	  buffer[count++] = okHOME;
	  buffer[count] = '\0';
	  focused->Input(buffer);
	  break;
	case XK_End: case XK_KP_End:
	  buffer[count++] = okEND;
	  buffer[count] = '\0';
	  focused->Input(buffer);
	  break;
	case XK_Page_Up: case XK_KP_Page_Up:
	  buffer[count++] = okPAGEUP;
	  buffer[count] = '\0';
	  focused->Input(buffer);
	  break;
	case XK_Page_Down: case XK_KP_Page_Down:
	  buffer[count++] = okPAGEDOWN;
	  buffer[count] = '\0';
	  focused->Input(buffer);
	  break;
	}
      }
    }
    break;

  //---------------------------------------------------------------------------
  case KeyRelease:
    {
      KeySym keysym;
      XComposeStatus compose;
      char buffer[40];
      int count = XLookupString((XKeyEvent*)event,buffer,39,&keysym,&compose);
      buffer[count]='\0';
      DBUG(D_WINDOW|D_LOW,"WINDOW[%p].KeyRelease(\"%s\")",this,buffer);
      if (focused) {
	if ((keysym >= XK_KP_Space) && (keysym<=XK_KP_9))
	  focused->ReleaseKey(buffer);
	else if ((keysym >= XK_space) && (keysym <= XK_asciitilde))
	  focused->ReleaseKey(buffer);
	else if ((keysym >= XK_nobreakspace) && (keysym <= XK_ydiaeresis))
	  focused->ReleaseKey(buffer);
	else switch (keysym) {
	case XK_Return: case XK_KP_Enter: case XK_Linefeed: case XK_Delete:
	case XK_BackSpace:
	  focused->ReleaseKey(buffer);
	  break;
	case XK_Left: case XK_KP_Left:
	  buffer[count++] = okLEFT;
	  buffer[count] = '\0';
	  focused->ReleaseKey(buffer);
	  break;
	case XK_Right: case XK_KP_Right:
	  buffer[count++] = okRIGHT;
	  buffer[count] = '\0';
	  focused->ReleaseKey(buffer);
	  break;
	case XK_Up: case XK_KP_Up:
	  buffer[count++] = okUP;
	  buffer[count] = '\0';
	  focused->ReleaseKey(buffer);
	  break;
	case XK_Down: case XK_KP_Down:
	  buffer[count++] = okDOWN;
	  buffer[count] = '\0';
	  focused->ReleaseKey(buffer);
	  break;
	case XK_Home: case XK_KP_Home:
	  buffer[count++] = okHOME;
	  buffer[count] = '\0';
	  focused->ReleaseKey(buffer);
	  break;
	case XK_End: case XK_KP_End:
	  buffer[count++] = okEND;
	  buffer[count] = '\0';
	  focused->ReleaseKey(buffer);
	  break;
	case XK_Page_Up: case XK_KP_Page_Up:
	  buffer[count++] = okPAGEUP;
	  buffer[count] = '\0';
	  focused->ReleaseKey(buffer);
	  break;
	case XK_Page_Down: case XK_KP_Page_Down:
	  buffer[count++] = okPAGEDOWN;
	  buffer[count] = '\0';
	  focused->ReleaseKey(buffer);
	  break;
	}
      }
    }
    break;

  //---------------------------------------------------------------------------
  default:
    WARN("WINDOW[%p].HandleEvent(%Xh) unknown event, see X.h",
	 this,event->type);
    break;
  }
}


void WINDOW::Refresh(int xa,int ya,int xb,int yb)
{
  IBUG(D_WINDOW|D_LOW,"WINDOW[%p].Refresh(%d,%d,%d,%d)",this,xa,ya,xb,yb);
  static FIGLET *f; 
  static ulong i;
  xb += xa-1; // Since xb,yb = width,height
  yb += ya-1;
  for (i = 1; (f = (FIGLET*)refresh[i]); ++i)
    if (f->Intersects(xa,ya,xb,yb)) f->Refresh();
  OBUG(D_WINDOW|D_LOW,"Done.");
}

//=============================================================================
// PEN
//=============================================================================

PEN::PEN(WINDOW *wnd,char *font_name)
  /// The creation of a pen, taking window and font it's supposed to use.
  /// After creating this, one should check wether pen->wnd() returns NULL.
  /// If pen->wnd() returns NULL, it means that the pen was not created for
  /// some reason, make sure you delete this pen, and don't use it, since
  /// it will guaranteed give you a fault with the X protocol.
{
  IBUG(D_PEN|D_HIGH,"new PEN[%p](WINDOW[%p],font=\"%s\")",this,wnd,font_name);
  ulong mask = 0;
  XGCValues values;
  window = wnd;
  font = XLoadQueryFont(window->dpy()->xdpy(),font_name);  // Allocate the font
  if (font == NULL) {       // Don't continue if the font couldn't be allocated
    window = NULL;
    WARN("new PEN[%p](\"%s\",\"%s\") cannot allocate font",
	 this,wnd->name(),font_name);
    OBUG(D_PEN|D_HIGH,"Cannot allocate font.");
    return;
  }
  DBUG(D_PEN|D_MED,"Done loading font[%p], creating GC",font);
  gc = XCreateGC(window->xdpy(),window->xwnd(),mask,&values);// Allocate the GC
  if (!gc) {                  // Don't continue if the gc couldn't be allocated
    XFreeFont(window->xdpy(),font);
    font = NULL;
    window = NULL;
    WARN("new PEN[%p](\"%s\",\"%s\") cannot create GC",
	 this,wnd->name(),font_name);
    OBUG(D_PEN|D_HIGH,"Cannot create GC.");
    return;
  }
  XSetFont(window->xdpy(),gc,font->fid);                   // Set the gc's font
  XSetForeground(window->xdpy(),gc,window->foreground());   // Foreground color
  XSetBackground(window->xdpy(),gc,window->background());   // Background color
  XSetLineAttributes(window->xdpy(),gc,1,LineSolid,CapProjecting,JoinMiter);
  XSetFillRule(window->xdpy(),gc,WindingRule);
  fontname = new char[strlen(font_name)+1];
  strcpy(fontname,font_name);
  window->Add(this);            // Register myself with the window that owns me
  OBUG(D_PEN|D_HIGH,"Created.");
}

PEN::~PEN(void)
  /// Releasing a pen from it's window, totally removing it.
  /// Since the pen is in no control of the ones who's using it, it cannot
  /// do anything to alert it's users of it's disappearance, the programmer
  /// should take care to make sure noone will use this pen hereafter.
{
  IBUG(D_PEN|D_HIGH,"~PEN[%p]()",this);
  if (window == NULL) {
    WARN("~PEN[%p](?,?) this pen was not initialised correctly!",this);
    OBUG(D_PEN|D_HIGH,"Not initialized.");
    return;
  }
  if (gc) XFreeGC(window->xdpy(),gc);      // Free the gc, if it is initialized
  if (font) XFreeFont(window->xdpy(),font);                // Free the font too
  if (fontname) delete [] fontname;
  window->Del(this);          // Deregister myself from the window that owns me
  OBUG(D_PEN|D_HIGH,"Deleted.");
}

//=============================================================================
