/*!f09!M-- OTOOLS.H --*/

/*
   NAME:	otools.h
   AUTHOR:	Ole Gunnar Westgaard TFU <olew@dina>
   DATE:	May 28 1996
   COMMENT:	Definitions for base otools classes
*/
#ifndef OTOOLSH
#define OTOOLSH

#include "list.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>

#include <stdio.h>
#include <string.h>

#define uchar unsigned char
#define uint  unsigned int
#define ulong unsigned long
#define ushort unsigned short

#include "debug.h"

// Text justification marks, upper 4 bits = alignement, lower 4 = rim
#define oMIDDLE 0x00
#define oBOTTOM 0x10
#define oTOP    0x20
#define oRIGHT  0x40
#define oLEFT   0x80
#define oHWIDE  0xC0
#define oVWIDE  0x30

// Settings for drawing arrows (PEN::Arrow())
#define oALEFT  0xC0
#define oARIGHT 0x80
#define oAUP    0x40
#define oADOWN  0x00
#define oAFLAT  0x20

// Standard color settings (colorset indexes)
#define ocBORDER     0
#define ocSHADOW     1
#define ocBACKGROUND 2
#define ocHIGHLIGHT  3
#define ocFOREGROUND 4
#define ocDARK       5
#define ocLIGHT      6
#define ocDISABLED   7
#define ocSELECTED   8

// ASCII / LATIN1 numbers for controllers
#define okHOME          0x01
#define okLEFT		0x02
#define okPAGEUP        0x03
#define okPAGEDOWN      0x04
#define okEND           0x05
#define okRIGHT		0x06
#define okBACKSPACE	0x08
#define okTAB		0x09
#define okENTER         0x0D
#define okDOWN		0x0E
#define okUP		0x10
#define okDELETE	0x7F
#define okPARAGRAPH	0xA7
#define	okCOPYRIGHT	0xA9
#define okREGISTERED	0xAE

class DISPLAY;
class COLORMAP;
class COLORSET;
class WINDOW;
class FIGLET;
class PEN;

class OTOOLS : public ELEMENT {
  /// There exists only one instance of this class in Otools, the user should
  /// never create another instance.
  ///
  /// This class enables the user to have windows open on several displays at
  /// once. Timing is also done by otools so that simple animation can be
  /// executed. Otools will call an elements Trigger (see class ELEMENT)
  /// when the timing is correct.
protected:
  LIST displays;
  struct TM { ELEMENT *element; ulong time,elapsed; TM *next; } *timer;
  struct FDS { int f; void (*c)(int); FDS *next; } *fdread,*fdwrite,*fdexcept;
  char running;
public:
  OTOOLS(void);
  //. Creates the Otools object. Is used only by Otools.
  char Add(DISPLAY*);
  //. When a class DISPLAY object is created, this will add itself to the
  //. Otools object with this method.
  char Del(DISPLAY*);
  //. Likeways, the display will remove itself from Otools when it is
  //. destroyed.
  char Timer(ELEMENT*,ulong);
  //. Set timer on element, 0=could not add
  char UnTime(ELEMENT*);
  //. Remove timer on element, 0=no such element
  char Release(ELEMENT*);
  //. Don't know ?
  char AddRead(int,void(*)(int));
  //. (not supported)
  char DelRead(int,void(*)(int));
  //. (not supported)
  char AddWrite(int,void(*)(int));
  //. (not supported)
  char DelWrite(int,void(*)(int));
  //. (not supported)
  char AddExcept(int,void(*)(int));
  //. (not supported)
  char DelExcept(int,void(*)(int));  
  //. (not supported)
  void EventLoop(void);
  //. Starts listening to all displays that have been Add()ed. Will not
  //. Return before there are no displays left, or the user issues Stop()
  void Stop(void);
  //. Stops the EventLoop().
};
/// }

extern OTOOLS otools;

class DISPLAY : public ELEMENT { // xdpy() == NULL if creation fails
  /// The class that keeps control of all displays in use.
  /// There are several ways to handle incoming events in this class, you
  /// should however refrain from using other Eventhandlers than
  /// class OTOOLS own EventLoop().
public:
                           DISPLAY(char*);
  //. create display
         virtual          ~DISPLAY(void);
  //. destructor
  inline virtual Window    xroot(void);
  //. X11 root window
  inline virtual Display  *xdpy(void);
  //. X11 display
  inline virtual int       xscr(void);
  //. screen of display
  inline virtual int       xfd(void);
  //. network filedescrip.
  inline virtual char     *name(void);
  //. displays name
  inline virtual int       Width(void);
  //. displays width
  inline virtual int       Height(void);
  //. displays height
  inline virtual COLORMAP *cmap(void);
  //. default colormap
  inline virtual void      Flush(void);
  //. flush display
  inline virtual void      Refresh(void);
  //. refresh windows
  inline virtual void      Stop(void);
  //. exit eventloop
         virtual void      EventLoop(void);
  //. enter eventloop
         virtual void      HandleCurrentEvents(void);
  //. handle buffer
         virtual void      HandleSingleEvent(void);
  //. handle next event
         virtual void      HandleSingleEvent(XEvent*);
  //. handle this event
         virtual void      NoWaitEvent(void);
  //. handle event
  inline virtual char      Add(COLORMAP*);
  //. add colormap
  inline virtual char      Del(COLORMAP*);
  //. del colormap
  inline virtual char      Add(WINDOW*);
  //. add window
  inline virtual char      Del(WINDOW*);
  //. remove window
  inline virtual char      Exists(WINDOW*);
  //. find a window
protected:
  Display *display;                                     // The X display
  char *title,running;                                  // displayname,running
  int screen,width,height;                              // screen and size
  COLORMAP *defcmap;                                    // default colormap
  LIST colormaps,windows;                               // lists
};

class COLORMAP : public ELEMENT {
  /// dpy() == NULL if creation fails
public:
  COLORMAP(DISPLAY*,Colormap);
  //. default colormap
  COLORMAP(DISPLAY*);
  //. new colormap
  virtual          ~COLORMAP(void);
  //. destructor
  inline virtual DISPLAY  *dpy(void);
  //. colormaps display
  inline virtual Colormap  xcmap(void);
  //. X11 colormap
  virtual ushort    Value(char*);
  //. hexadecimal -> int
  virtual char      AllocateColor(char*,ulong*);
  //. 1=succsess,0=fail
  virtual void      FreeColors(ulong*,int);
  //. free list of colors
  inline virtual char      Add(COLORSET*);
  //. remove colorset
  inline virtual char      Del(COLORSET*);
  //. remove colorset
protected:
  LIST colorsets;                                       // a list of colorsets
  Colormap colormap;                                    // X11 colormap
  DISPLAY *display;                                     // Colormaps display
};

class COLORSET : public ELEMENT {
public:
                            COLORSET(COLORMAP*,char*);
  //. Create colorset on a colormap, colors are generally separated by a
  //. colon. You can also use comma as a separator.
  //> new COLORSET(cmap,"red:gray40:gray60:gray80:black:blue:yellow");
  //. If creation of the colorset fails, cmap() will be NULL.
         virtual           ~COLORSET(void);
  //. Destructor, make sure no other classes are referencing this colorset.
  //. The colorset will release itself from the colormap it's owned by.
         virtual int        Add(char*);
  //. Add colors, -1 =fail
         virtual char       Change(int,char*);
  //. Change color, 0=fail
  inline virtual int        Size(void);
  //. Number of colors
  inline virtual char       Exist(int);
  //. Does index exist
  inline virtual ulong      operator[](int);
  //. Get color of index
  inline virtual COLORMAP  *cmap(void);
  //. Get the colormap it's made for
  inline virtual char      *name(int);
  //. Get color name of indexed color
protected:
  COLORMAP *colormap;                                   // it's colormap
  ulong    *set;                                        // color indexes
  char     *name_mem,**colors;                          // memory, colornames
  int       name_mem_size,count;                        // mem_size, num_colors
};
/// }

// Create() methods
#define WINDOWISMENU      0x01 // Set override_redirect

/// class WINDOW : public ELEMENT {
/// dpy() == NULL if creation fails

class WINDOW : public ELEMENT {
  /// The real meat of the window is the Create() function, all the builtin
  /// creators use this function to create a very flexible type of window.
  /// Take a close look at the function below (it's a protected member of
  /// window).

  //> char Create(DISPLAY *display,COLORMAP *colormap,
  //>             COLORSET *colorst,Window parent,char *title,
  //>             char *icontitle,int xpos,int ypos,
  //>             int min width,int min height,
  //>             int initial width,int initial height,
  //>             int max width,int max height,
  //>             int border width,uchar *icon,uint icon width,icon height,
  //>             char *fontname,int depth,
  //>             Visual *visual,uchar method);

  /// If parent=0 the window is not created as a child.
  ///
  /// If you set method=WINDOWISMENU, it will be created with no border.
  /// This is what is done in class SUBMENU.

  //> display->Add(this);
  //> XMapWindow(display->xdpy(),window);

  /// When you have created the window, add it to the display and map it:
  /// You're ready to go... For an example, have a look at test/xcat.

protected:
  Window    window;                                     // X11 window
  Pixmap    icon;                                       // X11 icon
  DISPLAY  *display;                                    // display of window
  COLORMAP *colormap;                                   // colormap of window
  COLORSET *colorset;                                   // colorset of window
  PEN      *pen;                                        // pen of window
  LIST      refresh,active,motion,pens;                 // various lists
  FIGLET   *focused,*selected;                          // keyboardfocus...
  int       bd_idx,bg_idx,fg_idx,                       // window color indexes
            width,height,posx,posy;                     // window size & pos
  char     *title;                                      // window title
  char      Create(DISPLAY *_dp,COLORMAP *_mp,          // display,colormap
		   COLORSET *_cs,Window _pt,char *_tt,  // colorset,parent,tt
		   char *_it,int _xp,int _yp,int miw,   // ititle,x,y,minwidth
		   int mih,int inw,int inh,int maw,     // sizes...
		   int mah,int _bw,uchar *_ic,uint _iw, // mah,bordw,icon...
		   uint _ih,char *_fn,int _dt,          // ... font,depth
		   Visual *_vs,uchar _mt);              // visual,method
public:
  WINDOW(void);
  //. does nothing, use Create() as described above
  WINDOW(COLORSET*,char*,char*,int,int,int,int);
  //. colorset,title,font,position,size
  WINDOW(COLORSET*,char*,char*,int,int,int,int,uchar*,uint,uint);
  //. colorset,title,font,position,size,icon
  WINDOW(uchar,WINDOW*,COLORSET*,char*,char*,
	 int,int,int,int, int,int,int,int,int,
	 uchar*,uint,uint,char*);
  //. mthd, parent, colorset, title, iconname, pos, sizes (min, ini, max),
  //. border, icon, font
         virtual          ~WINDOW(void);
  //. destroy window
  inline virtual DISPLAY  *dpy(void);
  //. the windows display
  inline virtual COLORMAP *cmap(void);
  //. windows colormap
  inline virtual COLORSET *cset(void);
  //. default colorset
  inline virtual PEN      *dpen(void);
  //. default pen
  inline virtual Display  *xdpy(void);
  //. X11 display
  inline virtual Window    xwnd(void);
  //. X11 window
  inline virtual char     *name(void);
  //. window name
  inline virtual int       Width(void);
  //. window width
  inline virtual int       Height(void);
  //. window height
  inline virtual int       xpos(void);
  //. window x position
  inline virtual int       ypos(void);
  //. window y position
  inline virtual ulong     border(void);
  //. border color
  inline virtual ulong     background(void);
  //. background color
  inline virtual ulong     foreground(void);
  //. foreground color
         virtual void      HandleEvent(XEvent*);
  //. incoming event
         virtual void      Refresh(int,int,int,int);
  //. refresh window area
  inline virtual void      Clear(void);
  //. clear window
  inline virtual void      Focus(FIGLET*);
  //. figlet request focus
  inline virtual void      DeFocus(FIGLET*);
  //. defocus figlet
  inline virtual char      Add(PEN*);
  //. add pen
  inline virtual char      Del(PEN*);
  //. remove pen
  inline virtual char      Add(FIGLET*);
  //. add to refresh
  inline virtual char      Del(FIGLET*);
  //. delete from refresh
  inline virtual char      Activate(FIGLET*);
  //. add to buttons
  inline virtual char      DeActivate(FIGLET*);
  //. delete from buttons
  inline virtual char      Motion(FIGLET*);
  //. add to motion
  inline virtual char      DeMotion(FIGLET*);
  //. delete from motion
  inline virtual void      Resize(int,int);
  //. resize window
  inline virtual void      LimitSize(int,int,int,int);
  //. window limit size
  inline virtual void      Enter(int,int);
  //. virtual enter
  inline virtual void      Leave(int,int);
  //. virtual leave
  inline virtual void      Move(void);
  //. window moved
  inline virtual void      Size(void);
  //. window resized
  inline virtual void      Before(int,int,int,int);
  //. virtual refresh
  inline virtual void      After(int,int,int,int);
  //. virtual refresh
};

class PEN : public ELEMENT {
protected:
  WINDOW *window;                           // The window that the pen draws in
  XFontStruct *font;                                               // gc's font
  GC gc;                                                         // The pens GC
  char *fontname;                  // The name of the font (allocated with new)
public:
  PEN(WINDOW*,char*);
  //. Create a pen with window & font, if it fails wnd() will be NULL.
  virtual ~PEN(void);
  //. Destroy this pen
  inline void Foreground(ulong);
  //. gc Foreground color
  inline void Background(ulong);
  //. gc Background color
  inline void LineAttributes(uint,int,int,int);
  //. gc Lineattributes
  inline WINDOW *wnd(void);
  //. Return the pointer to the pens WINDOW
  inline GC    xgc(void);
  //. The gc of the gc
  inline char  Font(char*);
  //. Attempt to change font, 0=failure
  inline char *Font(void);
  //. Return pointer to the pens fontname
  inline int   Descent(void);
  //. The height of font below line
  inline int   Ascent(void);
  //. The height of font above line
  inline int   FontHeight(void);
  inline int   Width(char*);
  //. The graphical with of a string
  inline void  Box(int,int,int,int);
  //. Draw a box
  inline void  Fbox(int,int,int,int);
  inline void  Frect(XRectangle*);
  inline void  Line(int,int,int,int);
  inline void  Clear(int,int,int,int);
  inline void  Write(int,int,char*);
  inline void  Write3d(int,int,char*,ulong,ulong,ulong);
  inline void  Fdisjoint(int,int,int,int,int,int,int,int,ulong);
  inline void  Justify(int*,int*,int,int,int,int,uchar,char*);
  inline void  Box3d(int,int,int,int,int,ulong,ulong);
  inline void  Fbox3d(int,int,int,int,int,ulong,ulong,ulong);
  inline void  Diamond(int,int,int,uchar,ulong,ulong);
  inline void  Diamond(int,int,int,int,uchar,ulong,ulong);
  inline void  Fdiamond(int,int,int,uchar,ulong,ulong,ulong);
  inline void  Fdiamond(int,int,int,int,uchar,ulong,ulong,ulong);
  inline void  Fpoly(XPoint*,int,ulong);
  inline void  Fcircle(int,int,int,int,ulong);
  inline void  Round(int,int,int,int,uchar,ulong);
  inline void  Fround(int,int,int,int,uchar,ulong,ulong,ulong);
  inline void  Clip(int,int,int,int);
  //. 0,0,0,0 removes clipmask
  inline void  Arrow(int,int,int,int,uchar,ulong,ulong,ulong,ulong);
};

// This class contains no data, except the WINDOW pointer, the figlet must
// belong to a window, if FIGLET.wnd() is NULL after creating the figlet, it
// means that the figlet was not created, and should be deleted. It also
// contains the functions that can be called by WINDOW upon Xwindow event.

class FIGLET : public ELEMENT {
protected:
  WINDOW *window;                             // The figlet belongs to a window
  int xa,ya,xb,yb;                                            // Size of figlet
public:
  inline FIGLET(WINDOW*);
  //. Creator, when you have your own Resize() function, wnd() is NULL when
  //. creation fails. With this you must set the size of the figlet yourself.
  inline FIGLET(WINDOW*,int,int,int,int);
  //. Creator, with size and builtin Resize(), see above.
  inline virtual ~FIGLET(void) {};
  inline virtual void Refresh(void);
  //. Force the figlet to be fully drawn
  inline virtual char Resize(int,int,int,int);
  //. resize, 0=success
  inline virtual char Intersects(int,int,int,int);
  //. wether box intersects
  inline virtual char Inside(int,int);
  //. wether point is inside
  inline virtual char Press(uchar,int,int);
  //. press, 1 to remain selected
  inline virtual char Motion(uchar,int,int);
  //. motion, 1 to remain selected
  inline virtual char Release(uchar,int,int);
  //. release, 1 to remain selected
  inline virtual char Focus(FIGLET*);
  //. get focus, 1 to remain in focus
  inline virtual void DeFocus(FIGLET*);
  //. forcibly loose focus
  inline virtual char Input(char*);
  //. keyboard input (latin1) 1 remains focus
  inline virtual char ReleaseKey(char*);
  //. keyboard release 1 remains in focus
  inline WINDOW *wnd(void);
  //. NULL means that figlet could not be created
};

#include "otools.i"

#endif /* OTOOLSH */
