/*
   NAME:	slider.cc
   AUTHOR:	Ole Gunnar Westgaard TFU <olew@dina>
   DATE:	Aug  2 1996
   COMMENT:	SLIDER, implements two figlets, ARROW & SLIDER
*/

#include "slider.h"

SLIDER::SLIDER(WINDOW *w,                         // window
	       int a,int b,int c,int d,           // size
	       char *n,                           // name
	       uchar tp,uchar fw,                 // type,framewidth
	       uchar wd,uchar ht,                 // width,height
	       uchar tr,                          // trim
	       float hv,float vv,                 // hor,ver
	       void(*cb)(SLIDER*,float,float))    // callback
  : FIGLET(w,a,b,c,d)
{
  inuse = 0;
  trim = tr;
  type = tp;
  thick = fw;
  callback = cb;
  width = wd;
  height = ht;
  ox = oy = 0;
  pen = window->dpen();
  cset = window->cset();
  if (n) {
    title = new char[strlen(n)+1];
    strcpy(title,n);
  } else title = NULL;
  if ((xb-xa) % 2) xb--;
  if ((yb-ya) % 2) yb--;
  Resize(xa,ya,xb,yb);
  Values(hv,vv);
  pen->Justify(&tx,&ty,xa,ya,xb,yb,trim,title);
  window->Add(this);
  window->Activate(this);
}

SLIDER::SLIDER(WINDOW *w,COLORSET *cs,PEN *pn,    // window,colorset,pen
	       int a,int b,int c,int d,           // size
	       char *n,                           // name
	       uchar tp,uchar fw,                 // type,framewidth
	       uchar wd,uchar ht,                 // width,height
	       uchar tr,                          // trim
	       float hv,float vv,                 // hor,ver
	       void(*cb)(SLIDER*,float,float))    // callback
  : FIGLET(w,a,b,c,d)
{
  inuse = 0;
  trim = tr;
  type = tp;
  thick = fw;
  callback = cb;
  width = wd;
  height = ht;
  ox = oy = 0;
  pen = pn ? pn : window->dpen();
  cset = cs ? cs : window->cset();
  if (n) {
    title = new char[strlen(n)+1];
    strcpy(title,n);
  } else title = NULL;
  if ((xb-xa) % 2) xb--;
  if ((yb-ya) % 2) yb--;
  Resize(xa,ya,xb,yb);
  Values(hv,vv);
  pen->Justify(&tx,&ty,xa,ya,xb,yb,trim,title);
  window->Add(this);
  window->Activate(this);
}

SLIDER::~SLIDER(void)
{
  window->DeActivate(this);
  window->Del(this);
  if (title) delete [] title;
}

float SLIDER::Hor(void)
{
  return hor;
}

float SLIDER::Ver(void)
{
  return ver;
}

void SLIDER::Values(float h,float v) /* Converts floats to mx,my */
{
  hor = h > 1.0 ? 1.0 : h < 0.0 ? 0.0 : h;
  ver = v > 1.0 ? 1.0 : v < 0.0 ? 0.0 : v;
  mx = (int)((float)(xb-xa-2*width) * hor);
  my = (int)((float)(yb-ya-2*height) * ver);
}

void SLIDER::Calculate(int mx,int my,int *a,int *b,int *c,int *d)
{
  int x = xa+width+mx, y = yb-height-my;
  /* oWHOR|oWVER = doesn't work as yet! */
  switch (type & 0xC0) {
  case oLEFT:   *a=xa;       *c=x+width; break;
  case oRIGHT:  *a=x-width;  *c=xb;      break;
  case oMIDDLE: *a=x-width;  *c=x+width; break;
  case oHWIDE:  *a=xa;       *c=xb;      break;
  }
  switch (type & 0x30) {
  case oTOP:    *b=ya;       *d=y+height; break;
  case oBOTTOM: *b=y-height; *d=yb;       break;
  case oMIDDLE: *b=y-height; *d=y+height; break;
  case oVWIDE:  *b=ya;       *d=yb;       break;
  }
  if ((type & 0xF0) == (oVWIDE|oHWIDE)) { // Special case
    int xm = (xa+xb+1)/2,
        ym = (ya+yb+1)/2;
    if (x<=xm) {
      *a = x-width;
      *c = xm+xm-x+width;
    } else {
      *a = xm+xm-x-width;
      *c = x+width;
    }
    if (y<=ym) {
      *b = y-height;
      *d = ym+ym-y+height;
    } else {
      *b = ym+ym-y-height;
      *d = y+height;
    }
  }
}

void SLIDER::Refresh(void)
{
  first = 1;
  pen->Foreground((*cset)[ocDARK]);
  pen->Fbox(xa,ya,xb,yb);
  Draw();
}

void SLIDER::Draw(void)
{
  int x1,y1,x2,y2,ax,ay,bx,by;

  Calculate(ox,oy,&ax,&ay,&bx,&by);
  Calculate(mx,my,&x1,&y1,&x2,&y2);
  pen->Fdisjoint(ax,ay,bx,by,x1,y1,x2,y2,(*cset)[ocDARK]);

  switch (type & 0x0E) {
  case sFLT:
    pen->Foreground((*cset)[type & 0x01 ? ocBACKGROUND : ocLIGHT]);
    pen->Fbox(x1,y1,x2,y2);
    break;
  case sSNG:
    if (type & 0x01)
      pen->Fbox3d(x1,y1,x2,y2,thick,
		  (*cset)[ocHIGHLIGHT],
		  (*cset)[ocBACKGROUND],
		  (*cset)[ocSHADOW]);
    else 
      pen->Fbox3d(x1,y1,x2,y2,thick,
		  (*cset)[ocSHADOW],
		  (*cset)[ocBACKGROUND],
		  (*cset)[ocHIGHLIGHT]);
    break;
  case sDBL:
    if (type & 0x01) {
      pen->Box3d(x1,y1,x2,y2,thick,(*cset)[ocHIGHLIGHT],(*cset)[ocSHADOW]);
      pen->Fbox3d(x1+thick,y1+thick,x2-thick,y2-thick,thick,
		  (*cset)[ocSHADOW],
		  (*cset)[ocBACKGROUND],
		  (*cset)[ocHIGHLIGHT]);
    } else {
      pen->Box3d(x1,y1,x2,y2,thick,(*cset)[ocSHADOW],(*cset)[ocHIGHLIGHT]);
      pen->Fbox3d(x1+thick,y1+thick,x2-thick,y2-thick,thick,
		  (*cset)[ocHIGHLIGHT],
		  (*cset)[ocBACKGROUND],
		  (*cset)[ocSHADOW]);
    }
    break;
  case sTRP:
    if (type & 0x01) {
      pen->Box3d(x1,y1,x2,y2,thick,(*cset)[ocHIGHLIGHT],(*cset)[ocSHADOW]);
      pen->Box3d(x1+thick,y1+thick,x2-thick,y2-thick,thick,
		 (*cset)[ocSHADOW],(*cset)[ocHIGHLIGHT]);
      pen->Fbox3d(x1+thick*2,y1+thick*2,x2-thick*2,y2-thick*2,thick,
		  (*cset)[ocHIGHLIGHT],(*cset)[ocBACKGROUND],
		  (*cset)[ocSHADOW]);
    } else {
      pen->Box3d(x1,y1,x2,y2,thick,(*cset)[ocSHADOW],(*cset)[ocHIGHLIGHT]);
      pen->Box3d(x1+thick,y1+thick,x2-thick,y2-thick,thick,
		 (*cset)[ocHIGHLIGHT],(*cset)[ocSHADOW]);
      pen->Fbox3d(x1+thick*2,y1+thick*2,x2-thick*2,y2-thick*2,thick,
		  (*cset)[ocSHADOW],
		  (*cset)[ocBACKGROUND],
		  (*cset)[ocHIGHLIGHT]);
    }
    break;
  case sLNE:
  default:
    if (type & 0x01)
      pen->Fbox3d(x1,y1,x2,y2,thick,
		  (*cset)[ocBACKGROUND],(*cset)[ocLIGHT],
		  (*cset)[ocBACKGROUND]);
    else
      pen->Fbox3d(x1,y1,x2,y2,thick,(*cset)[ocLIGHT],
		  (*cset)[ocBACKGROUND],(*cset)[ocLIGHT]);
    break;
  }
  if (title) {
    pen->Foreground((*cset)[ocFOREGROUND]);
    pen->Write(tx,ty,title);
  }
  ox = mx; oy = my;
  first = 0;
}

void SLIDER::Catch(int x,int y) // Convert screen x,y to scroll mx,my
{
  if (x < (xa+width)) x = xa+width;
  if (x > (xb-width)) x = xb-width;
  if (y < (ya+height)) y = ya+height;
  if (y > (yb-height)) y = yb-height;
  my = (yb-y-height);
  mx = (x-xa-width);
  hor = (float)((float)mx / (float)(xb-xa-2*width));
  ver = (float)((float)my / (float)(yb-ya-2*height));
}

char SLIDER::Press(uchar buttons,int x,int y)
{
  if (inuse) return 1;
  switch (buttons) {
  case 1: // select scrollbar
    Catch(x,y);
    if (((oy != my) || (ox != mx)) && (callback)) callback(this,hor,ver);
    inuse = 1;
    type ^= 0x01;
    Draw();
    return 1;
    break;
  }
  return 0;
}


char SLIDER::Motion(uchar buttons,int x,int y)
{
  if (!inuse) return 0;
  if (buttons && 0x01) {
    Catch(x,y);
    if (((oy != my) || (ox != mx)) && (callback)) callback(this,hor,ver);
    if (ox != mx || oy != my || first) Draw();
    return 1;
  }
  inuse = 0;
  type ^= 0x01;
  Draw();
  return 0;
}

char SLIDER::Release(uchar buttons,int x,int y)
{
  if (!inuse) return 0;
  switch (buttons) {
  case 1: // release scrollbar
    Catch(x,y);
    if (((oy != my) || (ox != mx)) && (callback)) callback(this,hor,ver);
    inuse = 0;
    type ^= 0x01;
    Draw();
    return 0;
    break;
  }
  return 1;
}

char SLIDER::Resize(int a,int b,int c,int d)
{
  FIGLET::Resize(a,b,c,d);
  Values(hor,ver);
  return 1;
}


